# -*- coding: cp1252 -*-

#    Script to calculate mean gridded temperature trends for GHCN subsets
#    Copyright (C) 2020 Dr. Ronan Connolly
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>
#
#    ********************************************************************
#
#    This script calculates station anomalies and gridded anomalies 
#    for GHCN version 3 and USHCN version 2 datasets
#    
#    Download the appropriate files from here:
#    From http://www.ncdc.noaa.gov/ghcnm/v3.php
#      GHCN v3 metadata file: ghcnm.tavg.v3.1.0.20120130.qcu.inv
#      GHCN v3 unadjusted monthly Tavg file: ghcnm.tavg.v3.1.0.20120130.qcu.dat
#      GHCN v3 adjusted monthly Tavg file: ghcnm.tavg.v3.1.0.20120130.qca.dat
#    From https://www.ncei.noaa.gov/data/us-historical-climatology-network/2.5/
#       ushcn-2.5.tavg.latest.FLs.52j.tar.gz
#       ushcn-2.5.tavg.latest.raw.tar.gz
#       ushcn-2.5.tavg.latest.tob.tar.gz
#    


# Load python libraries:
import matplotlib
matplotlib.use('Agg') # Use a non-interactive back-end to avoid plt.close() errors (non-critical) - this means you can't use plt.show()
import numpy as np
import os
import math

# input files and input file details. Modify these variables, if necessary
input_folder="input_files"
if not os.path.isdir(input_folder):
    print ("Cannot find "+input_folder+". Stopping analysis!")
    raise SystemExit

GHCN_version_num="v3.3.0"
GHCN_version_date="20190821"
GHCN_inv_file="ghcnm.tavg."+GHCN_version_num+"."+GHCN_version_date+".qcu.inv"
GHCN_unadj_file="ghcnm.tavg."+GHCN_version_num+"."+GHCN_version_date+".qcu.dat"
GHCN_adj_file="ghcnm.tavg."+GHCN_version_num+"."+GHCN_version_date+".qca.dat"
GHCN_folder=input_folder+"/ghcnm."+GHCN_version_num+"."+GHCN_version_date
USHCN_folder=input_folder+"/ushcn.v2.5.5.20190912"
USHCN_inv_file="ushcn-v2.5-stations.txt"

min_lat_considered=0
max_lat_considered=90
start_mon=0 #Jan
end_mon=11 #Dec
pre_1951_China_adjustment=0.09 # 0.09 for J-D, 0-90N; 0.14 for JJA, 30-70N
pre_1907_China_adjustment=0.28 # 0.28 for J-D, 0-90N; 0.16 for JJA, 30-70N
month_threshold=12 # number of months of data required per year

anom_start=1961
anom_end=1990
anom_threshold=15 # number of years of data required in anomaly period
QC=0 # set to 1 if you only want Quality Controlled temperatures

# Create output folders
trends_folder="output_temperature_trends"
if not os.path.isdir(trends_folder):
    os.mkdir(trends_folder)
general_folder="output_general"
if not os.path.isdir(general_folder):
    os.mkdir(general_folder)

# Create empty dictionaries
GHCN_info = {}
USHCN_info = {} # use to identify which GHCN stations are in the USHCN
used_stations={} # stations used in different subsets

grid_list={}
cur_lat=0
while cur_lat<36:
    grid_list[cur_lat]={}
    cur_lon=0
    while cur_lon<72:
        grid_list[cur_lat][cur_lon]=list()
        cur_lon=cur_lon+1
    cur_lat=cur_lat+1

# Read through USHCN inventory
if os.path.isfile(input_folder+"/"+USHCN_inv_file):
    USHCN_inv=open(input_folder+"/"+USHCN_inv_file,'r')
else:
    print ("Could not open "+USHCN_inv_file)
    raise SystemExit

for s in USHCN_inv:
    full_ID=s[0:11]
    ID=s[5:11]
    lat=float(s[12:20])
    lon=float(s[21:30])
    elev=s[31:37]
    state=s[38:40]
    name=s[41:70]
    comp1=s[72:78]
    comp2=s[79:85]
    comp3=s[86:92]
    time_zone=s[93:95]

    # Assign station details to info dictionary
    if lat>=min_lat_considered and lat<=max_lat_considered:
        USHCN_info[ID]={}
        USHCN_info[ID]["full_ID"]=full_ID
        USHCN_info[ID]["name"]=name
        USHCN_info[ID]["lat"]=lat
        USHCN_info[ID]["lon"]=lon
        USHCN_info[ID]["elev"]=elev
        USHCN_info[ID]["state"]=state
        # set flag for pre-1895 data to null
        USHCN_info[ID]["pre1895"]=0
USHCN_inv.close()

# Assign GHCN metadata
if os.path.isfile(GHCN_folder+"/"+GHCN_inv_file):
    GHCN_inv=open(GHCN_folder+"/"+GHCN_inv_file,'r')
else:
    print ("Could not open "+GHCN_inv_file)
    raise SystemExit

count=0
USHCN_count=0
# Read through each line in GHCN inventory file
for s in GHCN_inv:
    ID=s[0:11]
    country=int(s[0:3])
    if ID=="65103917000":
        # change the country code for the N. Irish (UK) Belfast/Alder
        # station to that for Ireland, so it gets included in our Irish construction
        country=621 
    USHCN_ID=s[5:11] #USHCN ID number
    lat=float(s[12:20])
    lon=float(s[21:30])
    elev=s[31:37]
    name=s[38:68].strip().replace("/","-").replace('"','-')
    # extract the two urbanization metrics from the GHCN metadata: RSU and ABC
    RSU=s[73:74] #Urbanisation (Population-based), R=rural, S=semi-urban, U=urban
    try:
        pop=int(s[75:79])
    except:
        pop=-1 #Some stations don't have population metadata, assume pop=-1, cause -9 already means <10,000
    ABC=s[106:107] #Urbanisation (Nightlights-based), A=rural, B=semi-urban, C=urban

    # we don't use the following metadata flags, but for reference this is where they are in the GHCN file
    #topography=s[79:81] # "FL"=flat, "HL"=hilly, "MT"=mountain top, "MV"= mountainous valley
    #vegetation=s[81:83] # "DE"=desert, "FO"=forest, "IC"=ice, "MA"=marsh, "xx"=other
    #water=s[83:85] # "CO" if coastal, "LA" if beside lake, otherwise "no"
    #airport=s[87:88] # "A" if it is an airport, otherwise "x"
    
    # Assign the metadata for this station to the GHCN_info dictionary (and USHCN_info if relevant)
    if lat>=min_lat_considered and lat<=max_lat_considered:
        GHCN_info[ID]={}
        GHCN_info[ID]["name"]=name
        GHCN_info[ID]["lat"]=lat
        GHCN_info[ID]["lon"]=lon
        GHCN_info[ID]["elev"]=elev
        GHCN_info[ID]["RSU"]=RSU
        GHCN_info[ID]["ABC"]=ABC
        # Convert GHCN's two urbanization metrics into our combined urbanization metric
        # "R", rural = rural in terms of both night brightness AND population
        # "U", urban = urban in terms of both night brightness AND population
        # "I", intermediate = all other stations
        if RSU=="R" and ABC=="A":
            GHCN_info[ID]["urban"]="R" # fully rural
        elif RSU=="U" and ABC=="C":
            GHCN_info[ID]["urban"]="U" # fully urban
        else:
            GHCN_info[ID]["urban"]="I" # intermediate urban
        GHCN_info[ID]["pop"]=pop
        GHCN_info[ID]["country"]=country
        if country==425 and USHCN_ID in USHCN_info:
            USHCN_count=USHCN_count+1
            GHCN_info[ID]["USHCN_ID"]=USHCN_ID
            USHCN_info[USHCN_ID]["GHCN_ID"]=ID
            USHCN_info[USHCN_ID]["GHCN_name"]=name
            USHCN_info[USHCN_ID]["RSU"]=RSU
            USHCN_info[USHCN_ID]["ABC"]=ABC
            USHCN_info[USHCN_ID]["pop"]=pop
            if USHCN_info[USHCN_ID]["RSU"]=="R" and USHCN_info[USHCN_ID]["ABC"]=="A":
                USHCN_info[USHCN_ID]["urban"]="R"
            elif USHCN_info[USHCN_ID]["RSU"]=="U" and USHCN_info[USHCN_ID]["ABC"]=="C":
                USHCN_info[USHCN_ID]["urban"]="U"
            else:
                USHCN_info[USHCN_ID]["urban"]="I"
        # Assign stations to grid boxes
        lat_num=int((lat+90.0)/5.0)
        lon_num=int((lon+180.0)/5.0)
        grid_list[lat_num][lon_num].append(ID)
        count=count+1

GHCN_inv.close()
print (str(count)+" GHCN stations identified")
print (str(USHCN_count)+" USHCN stations identified")

stats_file=open(general_folder+"/GHCN_station_statistics.dat", "w")
rural=0
urban=0
inter=0
for ID in GHCN_info:
    if GHCN_info[ID]["urban"]=="R":
        rural=rural+1 # fully rural station
    elif GHCN_info[ID]["urban"]=="U":
        urban=urban+1 # fully urban station
    else:
        inter=inter+1 # intermediately urban station
total=rural+urban+inter
stats_file.write("Global Network\n")
stats_file.write("Fully rural stations = "+str(rural)+"\t"+str(round(float(rural*100)/total,2))+"%\n")
stats_file.write("Intermediate stations = "+str(inter)+"\t"+str(round(float(inter*100)/total,2))+"%\n")
stats_file.write("Fully urban stations = "+str(urban)+"\t"+str(round(float(urban*100)/total,2))+"%\n")
stats_file.write("Total stations = "+str(total)+"\t100.00%\n")
stats_file.close()

# initialise earliest and latest year counters
year_start=9999
year_end=0


# Load Temperature data
print ("Loading GHCN data...")
GHCN_data={}

#Load GHCN data

#Load the raw (non-PHA adjusted) temperature records first.
print ("Loading Unadjusted GHCN data...")
try:
    unadj_file=open(GHCN_folder+"/"+GHCN_unadj_file,'r')
except:
    print ("Cannot open "+GHCN_unadj_file+". Closing...")
    raise SystemExit

for s in unadj_file:
    ID=s[0:11]
    year=int(s[11:15])
    
    year_avg=0.0
    num_months=0
    # if this station is in our list, then read in the temperature records
    if ID in GHCN_info:
        if not(ID in GHCN_data):
            GHCN_data[ID]={}
        x=19
        mon=0
        while mon<12:
            if mon>=start_mon and mon<=end_mon:
                # s[x:x+5]=monthly value in degC x 100
                # s[x+5:x+6]=DM flag; s[x+6:x+7]=QC flag; s[x+7:x+8]=source flag
                # load monthly value in degC & source
                # but first, check data is non-null and there is no QC flag (if requested):
                if s[x+6:x+7]==" " or QC!=1:
                    if int(s[x:x+5])!=-9999:
                        T=float(s[x:x+5])/100
                        
                        # Apply Valentia Observatory adjustments, if necessary
                        # Note we do not apply this non-climatic bias adjustment for the PHA-adjusted version
                        # since PHA is supposed to automatically detect any such biases (even though it doesn't)
                        # See Soon, Connolly and Connolly (2015). Re-evaluating the role of solar variability on Northern Hemisphere temperature trends since the 19th century. Earth-Science Reviews, 150, 409-452. https://doi.org/10.1016/j.earscirev.2015.08.010
                        if ID=="62103953000": # Valentia Observatory, Ireland's GHCN ID
                            if year>=2002:
                                T=T+0.3 # correct for Feb 2001 station move
                            elif year==2001 and mon>0:
                                T=T+0.3 # correct for Feb 2001 station move
                        
                        year_avg=year_avg+T
                        num_months=num_months+1
            x=x+8
            mon=mon+1
        if num_months>=month_threshold:
            GHCN_data[ID][year]={}
            GHCN_data[ID][year]["unadj_abs"]=round(year_avg/num_months,2)
            if year>year_start:
                year_start=year
            if year<year_end:
                year_end=year
unadj_file.close()

print ("Loading Adjusted GHCN data...")
try:
    adj_file=open(GHCN_folder+"/"+GHCN_adj_file,'r')
except:
    print ("Cannot open "+GHCN_adj_file+". Closing...")
    raise SystemExit

# Now, repeat process for the PHA-adjusted temperature records

for s in adj_file:
    ID=s[0:11]
    year=int(s[11:15])
    
    year_avg=0.0
    num_months=0
    if ID in GHCN_info:
        if not(ID in GHCN_data):
            GHCN_data[ID]={}
        x=19
        mon=0
        while mon<12:
            if mon>=start_mon and mon<=end_mon:
                # s[x:x+5]=monthly value in degC x 100
                # s[x+5:x+6]=DM flag; s[x+6:x+7]=QC flag; s[x+7:x+8]=source flag
                # load monthly value in degC & source
                # but first, check data is non-null and there is no QC flag (if requested):
                if s[x+6:x+7]==" " or QC!=1:
                    if int(s[x:x+5])!=-9999:
                        T=float(s[x:x+5])/100
                        
                        year_avg=year_avg+T
                        num_months=num_months+1
            x=x+8
            mon=mon+1
        if num_months>=month_threshold:
            if not(year in GHCN_data[ID]):
                GHCN_data[ID][year]={}
            GHCN_data[ID][year]["adj_abs"]=round(year_avg/num_months,2)
            if year>year_start:
                year_start=year
            if year<year_end:
                year_end=year
adj_file.close()



# Load USHCN temperature data
print ("Loading USHCN tob-adjusted data...")
USHCN_data={}

for ID in sorted(USHCN_info):
    USHCN_data[ID]={}
    file_prefix=USHCN_info[ID]["full_ID"]
    GHCN_ID=USHCN_info[ID]["GHCN_ID"]
    cur_filename=USHCN_folder+"/"+file_prefix+".tob.tavg"
    if os.path.isfile(cur_filename):
        cur_file=open(cur_filename, 'r')
        for k in cur_file:
            full_ID=k[0:11]
            year=int(k[12:16])

            year_avg=0.0
            num_months=0
            x=16
            mon=0
            while mon<12:
                if mon>=start_mon and mon<=end_mon:
                    try:
                        z=int(k[x:x+6])
                    except:
                        print (full_ID+"\t"+str(year)+str(mon)+"\t:"+k[x:x+5]+":\t"+"dataset")
                    DM_flag=k[x+6:x+7] # if not blank, i.e., "E" or "a-i" then it is an estimated value
                    QC_flag=k[x+7:x+8] # if not blank, i.e., "E","I","Q","X" then there are possible QC issues
                    DS_flag=k[x+8:x+9] # Data Source: blank = GHCN-Daily, otherwise different source
                    if int(k[x:x+6])!=-9999:
                        #monthly temperature
                        cur_temp=float(k[x:x+6])/100 # USHCN reported in deg. C x 100
                        year_avg=year_avg+cur_temp
                        num_months=num_months+1
                x=x+9
                mon=mon+1
            if num_months>=month_threshold:
                if year<year_start:
                    year_start=year
                if year>year_end:
                    year_end=year
                if not(year in GHCN_data[GHCN_ID]):
                    GHCN_data[GHCN_ID][year]={}
                GHCN_data[GHCN_ID][year]["tob_abs"]=round(year_avg/num_months,2)
        cur_file.close()
    else:
        print ("Could not open "+cur_filename)

# Load Temperature data
print ("Loading USHCN fully-adjusted data...")
USHCN_data={}

for ID in sorted(USHCN_info):
    USHCN_data[ID]={}
    file_prefix=USHCN_info[ID]["full_ID"]
    GHCN_ID=USHCN_info[ID]["GHCN_ID"]
    cur_filename=USHCN_folder+"/"+file_prefix+".FLs.52j.tavg"
    if os.path.isfile(cur_filename):
        cur_file=open(cur_filename, 'r')
        for k in cur_file:
            full_ID=k[0:11]
            year=int(k[12:16])

            year_avg=0.0
            num_months=0
            x=16
            mon=0
            while mon<12:
                if mon>=start_mon and mon<=end_mon:
                    try:
                        z=int(k[x:x+6])
                    except:
                        print (full_ID+"\t"+str(year)+str(mon)+"\t:"+k[x:x+5]+":\t"+"dataset")
                    DM_flag=k[x+6:x+7] # if not blank, i.e., "E" or "a-i" then it is an estimated value
                    QC_flag=k[x+7:x+8] # if not blank, i.e., "E","I","Q","X" then there are possible QC issues
                    DS_flag=k[x+8:x+9] # Data Source: blank = GHCN-Daily, otherwise different source
                    if int(k[x:x+6])!=-9999:
                        #monthly temperature
                        cur_temp=float(k[x:x+6])/100 # USHCN reported in deg. C x 100
                        year_avg=year_avg+cur_temp
                        num_months=num_months+1
                x=x+9
                mon=mon+1
            if num_months>=month_threshold:
                if year<year_start:
                    year_start=year
                if year>year_end:
                    year_end=year
                if not(year in GHCN_data[GHCN_ID]):
                    GHCN_data[GHCN_ID][year]={}
                GHCN_data[GHCN_ID][year]["FLs_52j_abs"]=round(year_avg/num_months,2)
        cur_file.close()
    else:
        print ("Could not open "+cur_filename)

###############
# Anomalies
###############

print ("Calculating temperature anomaly records for each station...")

for ID in sorted(GHCN_data):
    first_year=2100
    last_year=0
    unadj_lor=0 #length of record
    for year in GHCN_data[ID]:
        if "unadj_abs" in GHCN_data[ID][year]:
            if year<first_year:
                first_year=year
            if year>last_year:
                last_year=year
            unadj_lor=unadj_lor+1
    GHCN_info[ID]["unadj_lor"]=unadj_lor
    GHCN_info[ID]["unadj_first"]=first_year
    GHCN_info[ID]["unadj_last"]=last_year

# Calculate and output temperature anomalies for each station
for ID in sorted(GHCN_data):
    unadj_Ts=[]
    adj_Ts=[]
    tob_Ts=[]
    FLs_52j_Ts=[]
    year=anom_start
    while year<=anom_end:
        if year in GHCN_data[ID]:
            if "unadj_abs" in GHCN_data[ID][year]:
                unadj_Ts.append(GHCN_data[ID][year]["unadj_abs"])
            if "tob_abs" in GHCN_data[ID][year]:
                tob_Ts.append(GHCN_data[ID][year]["tob_abs"])
            if "adj_abs" in GHCN_data[ID][year]:
                adj_Ts.append(GHCN_data[ID][year]["adj_abs"])
            if "FLs_52j_abs" in GHCN_data[ID][year]:
                FLs_52j_Ts.append(GHCN_data[ID][year]["FLs_52j_abs"])
        year=year+1
    if len(unadj_Ts)>=anom_threshold:
        unadj_anom_mean=round(np.mean(unadj_Ts),2)
        GHCN_info[ID]["unadj_anom_mean"]=unadj_anom_mean
        for year in GHCN_data[ID]:
            if "unadj_abs" in GHCN_data[ID][year]:
                GHCN_data[ID][year]["unadj_anom"]=GHCN_data[ID][year]["unadj_abs"]-unadj_anom_mean
    if len(tob_Ts)>=anom_threshold:
        tob_anom_mean=round(np.mean(tob_Ts),2)
        GHCN_info[ID]["tob_anom_mean"]=tob_anom_mean
        for year in GHCN_data[ID]:
            if "tob_abs" in GHCN_data[ID][year]:
                GHCN_data[ID][year]["tob_anom"]=GHCN_data[ID][year]["tob_abs"]-tob_anom_mean
    if len(adj_Ts)>=anom_threshold:
        adj_anom_mean=round(np.mean(adj_Ts),2)
        GHCN_info[ID]["adj_anom_mean"]=adj_anom_mean
        for year in GHCN_data[ID]:
            if "adj_abs" in GHCN_data[ID][year]:
                GHCN_data[ID][year]["adj_anom"]=GHCN_data[ID][year]["adj_abs"]-adj_anom_mean
    if len(FLs_52j_Ts)>=anom_threshold:
        FLs_52j_anom_mean=round(np.mean(tob_Ts),2)
        GHCN_info[ID]["FLs_52j_anom_mean"]=FLs_52j_anom_mean
        for year in GHCN_data[ID]:
            if "FLs_52j_abs" in GHCN_data[ID][year]:
                GHCN_data[ID][year]["FLs_52j_anom"]=GHCN_data[ID][year]["FLs_52j_abs"]-FLs_52j_anom_mean



###############
# Gridding
###############

# Convert station anomalies into gridded anomalies
print ("Calculating gridded averages...")
region_combinations=[(0,0,0,1,0),(0,0,1,0,0),(0,1,0,0,0),(1,0,0,0,0),
                    (1,1,0,0,0),(1,0,1,0,0),(1,0,0,1,0),
                    (0,1,1,0,0),(0,1,0,1,0),(0,0,1,1,0),
                    (1,1,1,0,0),(1,1,0,1,0),(1,0,1,1,0),(0,1,1,1,0),
                    (1,1,1,1,0),(0,0,0,0,1)]
for sampling in ["Rural_only","Urban_or_rural"]:
    print (sampling)
    # Arctic, Ireland, China, USA, everywhere
    for regions_considered in region_combinations:
        Arctic=0
        Ireland=0
        China=0
        USA=0
        everywhere=0
        subset_name=""
        num_regions=0
        if regions_considered[0]==1:
            Arctic=1
            subset_name=subset_name+"_Arctic"
            num_regions=num_regions+1
        if regions_considered[1]==1:
            Ireland=1
            subset_name=subset_name+"_Ireland"
            num_regions=num_regions+1
        if regions_considered[2]==1:
            China=1
            subset_name=subset_name+"_China"
            num_regions=num_regions+1
        if regions_considered[3]==1:
            USA=1
            subset_name=subset_name+"_USA"
            num_regions=num_regions+1
        if regions_considered[4]==1:
            everywhere=1
            subset_name=subset_name+"_All_NH"
            num_regions=num_regions+1
        print ("Region combination:\t"+subset_name)
    
        if everywhere==0:
            grid_file=open(trends_folder+"/gridded_"+sampling+"_"+str(num_regions)+"_regions"+subset_name+".txt","w")
        else:
            grid_file=open(trends_folder+"/gridded_"+sampling+"_all_regions"+subset_name+".txt","w")
        grid_file.write(sampling+"\n"+str(num_regions)+" regions\n"+subset_name+"\n")
        grid_file.write("Temperature anomalies (deg C) ")
        grid_file.write("Relative to "+str(anom_start)+"-"+str(anom_end)+"\n")
        grid_file.write("Year\tT\t#Stations\t#Rural\t#Inter\t#Urban\n")

        if everywhere==0:
            stations_file=open(general_folder+"/stations_used_"+sampling+"_"+str(num_regions)+"_regions"+subset_name+".txt","w")
        else:
            stations_file=open(general_folder+"/stations_used_"+sampling+"_all_regions"+subset_name+".txt","w")
        stations_file.write(sampling+"\n"+str(num_regions)+" regions\n"+subset_name+"\n")
        stations_file.write("Lat\tLon\tID\tUrban\tLOR\tFirst\tLast\tCountry\tName\n")

    
        used_stations[subset_name]={}
        annual_anom={}
        year=1800
        while year<=2018:
            annual_anom[year]={}
            station_count=0
            rural_count=0
            inter_count=0
            urban_count=0
            curT_anom=[]
            cur_lat_w_anom=[]
            cur_grid_w_anom=[]
            cur_w_i_anom=[]
            # Iterate through all gridboxes with stations
            cur_lat=0
            while cur_lat<36:
                lat_weight=math.cos(math.radians(abs((cur_lat*5.0)-90.0 + 2.5)))
                # 0...5=unadj, tob, adj, tob_adjust, step_adjust, all_adjust
                cur_lon=0
                while cur_lon<72:
                    # 0...5=unadj, tob, adj, tob_adjust, step_adjust, all_adjust
                    num_stations=0
                    for ID in sorted(grid_list[cur_lat][cur_lon]):
                        if ID in GHCN_data:
                            if year in GHCN_data[ID]:
                                meets_subset_conditionals=0
                                if sampling=="Rural_only":
                                    # Make sure this ID is eligible for the analysis
                                    if Arctic==1:
                                        if GHCN_info[ID]["lat"]>=60 and GHCN_info[ID]["urban"]=="R":
                                            meets_subset_conditionals=1
                                    if Ireland==1:
                                        if GHCN_info[ID]["country"]==621 and GHCN_info[ID]["urban"]=="R":
                                            meets_subset_conditionals=1
                                    if China==1:
                                        if GHCN_info[ID]["country"]==205:
                                            #figure out relevant urban combination
                                            if year>1990:
                                                if GHCN_info[ID]["urban"]=="R" or GHCN_info[ID]["urban"]=="I":
                                                    meets_subset_conditionals=1
                                            elif year>=1951:
                                                if GHCN_info[ID]["urban"]=="R":
                                                    meets_subset_conditionals=1
                                            elif year>=1907:
                                                if GHCN_info[ID]["urban"]=="R" or GHCN_info[ID]["urban"]=="I":
                                                    meets_subset_conditionals=1
                                            else:
                                                meets_subset_conditionals=1
                                    if USA==1:
                                        if GHCN_info[ID]["country"]==425 and GHCN_info[ID]["lat"]>24 and GHCN_info[ID]["lat"]<50 and GHCN_info[ID]["urban"]=="R":
                                            meets_subset_conditionals=1
                                    if everywhere==1:
                                        if GHCN_info[ID]["urban"]=="R":
                                            meets_subset_conditionals=1
                                        
                                elif sampling=="Urban_or_rural":
                                    # then urban and rural stations are included
                                    if Arctic==1:
                                        if GHCN_info[ID]["lat"]>=60:
                                            meets_subset_conditionals=1
                                    if Ireland==1:
                                        if GHCN_info[ID]["country"]==621:
                                            meets_subset_conditionals=1
                                    if China==1:
                                        if GHCN_info[ID]["country"]==205:
                                            meets_subset_conditionals=1
                                    if USA==1:
                                        if GHCN_info[ID]["country"]==425 and GHCN_info[ID]["lat"]>24 and GHCN_info[ID]["lat"]<50:
                                            meets_subset_conditionals=1
                                    if everywhere==1:
                                        meets_subset_conditionals=1
    
                                if meets_subset_conditionals==1:
                                    if sampling=="Rural_only":
                                        if GHCN_info[ID]["country"]==425:
                                            # contiguous United States.
                                            # Use TOB adjusted-data to correct for documented changes in Time of OBservation biases
                                            if "tob_anom" in GHCN_data[ID][year]:
                                                curT_anom.append(GHCN_data[ID][year]["tob_anom"])
                                                cur_lat_w_anom.append(lat_weight)
                                                num_stations=num_stations+1
                                                station_count=station_count+1
                                                if GHCN_info[ID]["urban"]=="R":
                                                    rural_count=rural_count+1
                                                elif GHCN_info[ID]["urban"]=="U":
                                                    urban_count=urban_count+1
                                                else:
                                                    inter_count=inter_count+1
                                                if not(ID in used_stations[subset_name]):
                                                    used_stations[subset_name][ID]={}
                                        elif GHCN_info[ID]["country"]==205:
                                            # China. Apply urbanization bias corrections if necessary. 
                                            # See Soon, Connolly and Connolly (2015). Re-evaluating the role of solar variability on Northern Hemisphere temperature trends since the 19th century. Earth-Science Reviews, 150, 409-452. https://doi.org/10.1016/j.earscirev.2015.08.010
                                            if year>=1951:
                                                China_adjustment=0.0
                                            elif year>=1907:
                                                China_adjustment=pre_1951_China_adjustment
                                            else:
                                                China_adjustment=pre_1907_China_adjustment
                                            if "unadj_anom" in GHCN_data[ID][year]:
                                                curT_anom.append(GHCN_data[ID][year]["unadj_anom"]+China_adjustment)
                                                cur_lat_w_anom.append(lat_weight)
                                                num_stations=num_stations+1
                                                station_count=station_count+1
                                                if GHCN_info[ID]["urban"]=="R":
                                                    rural_count=rural_count+1
                                                elif GHCN_info[ID]["urban"]=="U":
                                                    urban_count=urban_count+1
                                                else:
                                                    inter_count=inter_count+1
                                                if not(ID in used_stations[subset_name]):
                                                    used_stations[subset_name][ID]={}
                                        else:
                                            # Any other country
                                            if "unadj_anom" in GHCN_data[ID][year]:
                                                curT_anom.append(GHCN_data[ID][year]["unadj_anom"])
                                                cur_lat_w_anom.append(lat_weight)
                                                num_stations=num_stations+1
                                                station_count=station_count+1
                                                if GHCN_info[ID]["urban"]=="R":
                                                    rural_count=rural_count+1
                                                elif GHCN_info[ID]["urban"]=="U":
                                                    urban_count=urban_count+1
                                                else:
                                                    inter_count=inter_count+1
                                                if not(ID in used_stations[subset_name]):
                                                    used_stations[subset_name][ID]={}
                                    else:
                                        # All stations urban or rural considered
                                        # and using PHA-adjusted GHCN (see Menne & Williams (2009) for details on the automated PHA process)
                                        if "FLs_52j_anom" in GHCN_data[ID][year]:
                                            # This is a USHCN station
                                            # So, use fully homogenized USHCN dataset if available
                                            curT_anom.append(GHCN_data[ID][year]["FLs_52j_anom"])
                                            cur_lat_w_anom.append(lat_weight)
                                            num_stations=num_stations+1
                                            station_count=station_count+1
                                            if GHCN_info[ID]["urban"]=="R":
                                                rural_count=rural_count+1
                                            elif GHCN_info[ID]["urban"]=="U":
                                                urban_count=urban_count+1
                                            else:
                                                inter_count=inter_count+1
                                            if not(ID in used_stations[subset_name]):
                                                used_stations[subset_name][ID]={}
                                        elif "adj_anom" in GHCN_data[ID][year]:
                                            # This is not a USHCN station
                                            # So, use PHA-adjusted GHCN dataset
                                            curT_anom.append(GHCN_data[ID][year]["adj_anom"])
                                            cur_lat_w_anom.append(lat_weight)
                                            num_stations=num_stations+1
                                            station_count=station_count+1
                                            if GHCN_info[ID]["urban"]=="R":
                                                rural_count=rural_count+1
                                            elif GHCN_info[ID]["urban"]=="U":
                                                urban_count=urban_count+1
                                            else:
                                                inter_count=inter_count+1
                                            if not(ID in used_stations[subset_name]):
                                                used_stations[subset_name][ID]={}
                    if num_stations>0:
                        z=0
                        while z<num_stations:
                            cur_grid_w_anom.append(num_stations)
                            z=z+1
                    cur_lon=cur_lon+1
                cur_lat=cur_lat+1
    
            grid_file.write(str(year)+"\t")
            global_T=0.0
            w_i=[]
            for i,val in enumerate(curT_anom):
                w_i.append(cur_lat_w_anom[i]/cur_grid_w_anom[i])
            sum_w_i=np.sum(w_i)
            for i,val in enumerate(curT_anom):
                global_T=global_T+(curT_anom[i]*w_i[i])
            if sum_w_i>0.0:
                global_T=round(global_T/sum_w_i,2)
                grid_file.write(str(global_T))
                annual_anom[year]=global_T
    
            grid_file.write("\t"+str(station_count)+"\t"+str(rural_count))
            grid_file.write("\t"+str(inter_count)+"\t"+str(urban_count)+"\n")
    
            year=year+1
    
        station_count=0
        rural_count=0
        inter_count=0
        urban_count=0
        lats=[]
        for ID in used_stations[subset_name]:
            station_count=station_count+1
            if GHCN_info[ID]["urban"]=="R":
                rural_count=rural_count+1
            elif GHCN_info[ID]["urban"]=="U":
                urban_count=urban_count+1
            else:
                inter_count=inter_count+1
            lats.append(GHCN_info[ID]["lat"])
            stations_file.write(str(GHCN_info[ID]["lat"])+"\t")
            stations_file.write(str(GHCN_info[ID]["lon"])+"\t")
            stations_file.write(str(ID)+"\t"+GHCN_info[ID]["urban"]+"\t"+str(GHCN_info[ID]["unadj_lor"])+"\t"+str(GHCN_info[ID]["unadj_first"])+"\t"+str(GHCN_info[ID]["unadj_last"])+"\t"+str(GHCN_info[ID]["country"])+"\t"+GHCN_info[ID]["name"]+"\n")
    
        grid_file.write("\nTotal\t\t"+str(station_count)+"\t"+str(rural_count))
        grid_file.write("\t"+str(inter_count)+"\t"+str(urban_count)+"\n")
    
        if len(lats)>0:
            mean_lat=np.mean(lats)
            min_lat=min(lats)
            max_lat=max(lats)
            cos_lat=np.cos(math.radians(mean_lat))
                            
            grid_file.write("\nMin lat =\t"+str(min_lat)+" deg N\n")
            grid_file.write("Max lat =\t"+str(max_lat)+" deg N\n")
            grid_file.write("Mean lat=\t"+str(round(mean_lat,2))+" deg N\n")
            grid_file.write("Cos(Mean lat)=\t"+str(round(cos_lat,3))+"\n")
        else:
            print (subset_name, "no stations")
    
        grid_file.close()

        stations_file.close()
    
print ("Done.")